# Oberski, D. & A. Satorra (forthcoming). 
# http://daob.org/ 

# Author: Daniel Oberski
# Date: 19 April 2010
# License: GPL >= 2 


get.var.corrected <- function(mxFit, fixed.params, V.fixed, 
        Gamma=NULL, W=NULL, extra=FALSE, uncorrected=TRUE)  {
    
    if(!all(dim(V.fixed) == c(length(fixed.params), length(fixed.params))))
        stop("The variance matrix of the fixed parameters should be square with the same number of rows and columns as the number of fixed parameters.")

    nobs <- mxFit@data@numObs
    S <- Matrix(mxFit@data@observed) # Observed covariance matrix
    m <- get.LISREL.matrices.from.RAM(mxFit) # Matrices transformed to LISREL
    Delta <- dvec.Sigma.LISREL(m$LY, m$PH, m$BE, m$PS) 
    free.params <- get.mx.free.indices(mxFit)

    D.free <- Delta[, free.params]
    D.fixed <- Delta[, fixed.params]

    # If 4th order moments are not given, calculate them assuming normality.
    if (is.null(Gamma)) { 
        Gamma <- get.gamma.star(S)
    }
    # If weights not given, assume normal-theory Maximum Likelihood was used to
    #   estimate the model.
    if (is.null(W)) {
        Sigma <- mxExpected(mxFit)     # Expected covariance matrix
        W <- get.gamma.star.inv(Sigma) # Weights used in the estimation
    }

    J <- t(D.free) %*% W %*% D.free
    Ji <- solve(J)
    B <- t(D.free) %*% W %*% Gamma %*% W %*% D.free

    # Derivatives wrt free parameters, then wrt fixed params:
    Dd.fix <- t(D.free) %*% W %*% D.fixed


    if (extra)  {
    # Only output what should be added to the variance to obtain
    #   correct standard errors
        # V := (1/2 H)^-1 and V = n * J so
        E <-  (Dd.fix %*% V.fixed %*% t(Dd.fix)) 
        avar.theta <- (Ji %*% E %*% t(Ji)) 
    }
    else { # Just output corrected variance
        E <-  (Dd.fix %*% V.fixed %*% t(Dd.fix))* nobs
        avar.theta <- (Ji %*% (B + E) %*% t(Ji)) / nobs
    }

    if(uncorrected)
        return (list('corrected' = avar.theta, 
                     'uncorrected' =  (Ji %*% B %*% t(Ji))/nobs ))

    return(avar.theta)
}


get.stderr.corrected <- function(mxFit, fixed.params, V.fixed, 
        Gamma=NULL, W=NULL, extra=FALSE, DEBUG=FALSE) {
    avar.theta <- get.var.corrected(mxFit, fixed.params, V.fixed, Gamma, W, 
            extra, DEBUG)
    
    sd.theta <- sqrt(diag(avar.theta))

    return(sd.theta)
}

# Given a matrix name, mxModel object, and row and column index of 
#  the matrix (e.g. "A", 2, 3), return the parameter index from the
#  LISREL model parameter vector := [vec F, vech S, vec A, 0]
get.par.index <- function(matname, row, col, mxFit) {
    if(! matname %in% c('S', 'A'))
        stop('Matname should be either "S" or "A".')

    ny <- nrow(mxFit@matrices$F@values)
    ne <- ncol(mxFit@matrices$F@values)

    if (matname == 'A') {# LY (F) and vech PH (vech S) come before
        offset <-  (ny * ne) + ((ne * (ne + 1))/2) 
        flattened.index <- (ne-1) * col  +  row
    }
    else { # Only LY (F) comes before
        offset <-  (ny * ne) 
        flattened.index <- get.flattened.vech.index(ne, row, col)
    }

    return(offset + flattened.index)
}
