# Some functions useful in matrix algebra. duplication.matrix, commutation.matrix, 
#  and elimination.matrix are taken from matrixcalc 
#  (but redone to work with the Matrix package).
#  Author: Daniel Oberski
#  Date: 19 April 2010
#  License: same as R (GPL version 2): http://www.r-project.org/COPYING

# Stack all elements column-wise
vec <- as.numeric

# stack non-upper-triangular elements column-wise
vech <- function (x) { 
    return(t(t(x[!upper.tri(x)])))
}


Kmat <- function(m, n) {
# This version of commutation.matrix has 3 advantages over the
#  version encountered in matrixcalc:
#    1. Much faster
#    2. Consumes less memory by taking advantage of sparse Matrix
#    3. Outputs a Matrix, allowing Matrix package to take further
#        advantage of sparseness is subsequent multiplications etc.
    p <- m*n
    Ki <- rep(0, p)
    K <- Matrix(0, p, p, sparse=TRUE)
    res <- .C("commutation_indices", as.integer(m), as.integer(n), Ki=Ki)
    K[res$Ki] <- 1
    K
}


# From package matrixcalc
duplication.matrix <- function (n = 1) 
{
    if ((n < 1) | (round(n) != n)) 
        stop("n must be a positive integer")
    d <- Matrix(0, n * n, n * (n + 1)/2)
    count = 0
    for (j in 1:n) {
        d[(j - 1) * n + j, count + j] = 1
        if (j < n) {
            for (i in (j + 1):n) {
                d[(j - 1) * n + i, count + i] <- 1
                d[(i - 1) * n + j, count + i] <- 1
            }
        }
        count = count + n - j
    }
    return(d)
}

elimination.matrix  <- function (n) 
{
    if (n <= 0) 
        stop("argument n is not positive")
    if (n != trunc(n)) 
        stop("argument n is not an integer")
    A <- Matrix(0, nrow = n, ncol = n)
    for (i in 1:n) for (j in 1:n) A[i, j] <- i + j
    vechA <- vech(A)
    vecA <- vec(A)
    q <- n * n
    p <- n * (n + 1)/2
    D <- matrix(0, nrow = p, ncol = q)
    for (i in 1:p) {
        value.set <- FALSE
        for (j in 1:q) {
            if (!value.set) {
                D[i, j] <- as.numeric(vecA[j, 1] == vechA[i, 
                  1])
                if (D[i, j] == 1) 
                  value.set <- TRUE
            }
        }
    }
    return(D)
}

# Moore-penrose inverse of matrix X (M&N p.258)
mp.inv <- function(X) {
  svd.X <- svd(X)
  T <- svd.X$v
  S <- svd.X$u
  L.i <- diag(1/svd.X$d)

  T %*% L.i %*% t(S)
}


# convenience function that gives
#   an identity Matrix of order p
Id <- function(p) {Matrix(diag(rep(1, p)))}

# Selection matrices E and Tij (Neudecker and Satorra 1991)
get.E <- function(i, j, p) {
  Ip <- Id(p)
  Ip[,i] %*% t(Ip[,j])
}

get.T <- function(i, j, p) {
  get.E(i, j, p) + get.E(j, i, p)
}


get.symm.diag <- function(p) {
  # Return the indices for the diagonal
  #  of a vech-torized symmetric matrix
  Vectorize(get.i <- function(p, i) {
    if(i == 1) return(i)
    get.i(p, i-1) + p - (i - 2)
  })(p, 1:p)
}

# Given <nrows> := no. nonredundant elements at the column <col>,
# and <row> is the row number
#  return the flattened index of the [row, col] element of vech
get.vech.length <- function(nrow, row, col) {
    if (col == 1) return(row)
    get.vech.length(nrow + 1, nrow, col - 1) + row
}

# Given the number of rows and a [row, col] index of a full
# matrix (say, S), Return the index number of vech(S) needed
# to retrieve that same element.
get.flattened.vech.index <- function(nrow, row, col) {
    if (col > row) {
        tmp <- row
        row <- col
        col <- tmp
    }
    get.vech.length(nrow - col + 1, row, col)
}


