# Various helper functions that take an MxFit or model object as input

# Author: Daniel Oberski
# Date: 19 April 2010
# License: GPL >= 2 



# Expected (implied) covariance matrix for the RAM model
#  Note that RAM and LISREL models can be written in
#  terms of each other.
mxExpected <- function(mxObj, filter=TRUE) {
  A <- mxObj@matrices$A@values 
  S <- mxObj@matrices$S@values
  F <- mxObj@matrices$F@values # "Filter" (elimination) matrix

  eye <- solve(Id(nrow(S)) - A) # reduced form directed paths inverse
  vEE <- eye %*% S %*% t(eye)
  if (filter)
      return(F %*% vEE %*% t(F)) # see McArdle 1980
  else
      return(vEE) # see McArdle 1980
}

# A function that gets matrices from openmx via the
#   estimates vector theta. This allows the numerical
#   derivatives function to re-adjust the theta and get
#   a different implied covariance matrix.
RAMGetMatrix <- function(mxObj, name) {
  # Indicator matrices of free or fixed parameters
  A.free <- mxObj@matrices$A@free
  S.free <- mxObj@matrices$S@free

  theta <- mxObj@output$estimate
  mat <-  mxObj@matrices[[name]]@values

  # This uses that A and S are the only matrices that
  #  contain free parameters, and A is always first in 
  # i.e. theta := (vec A, vech S)
  # Let's hope that behavior is not changed without warning...
  if (name == 'A') {  
    mat[A.free] <- theta[1:sum(A.free)]
  }
  else if (name == 'S') {
    mat[S.free] <- theta[(sum(A.free)+1):length(theta)]
  }
  mat
}

# McArdle & McDonald's RAM model can be reformulated into LISREL all-y
#  model. This is just done for reading and referenntial ease with respect
#  the various articles of Satorra cum suis, who  use the 
#  LISREL rather than the RAM model. 
# Matrix classes from the Matrix package are returned.
get.LISREL.matrices.from.RAM <- function(mxFit) {
    mats <- list()

    LY <- Matrix(mxEval(F, mxFit))
    PH <- Matrix(mxEval(S, mxFit))
    BE.0 <- Matrix(mxEval(A, mxFit))
    BE <- Matrix(Id(ncol(BE.0)) - BE.0)  # used in all results
    PS <- Matrix(0, p <- nrow(LY), p)

    list(LY = LY, PH = PH, BE.0 = BE.0, BE = BE, PS = PS)
}

# Return matrix flattened indices of the free parameter of the model,
#  assuming the model formulation of Neudecker and Satorra 1991 (first page)
get.mx.free.indices <- function(mxFit) {
    # Filter matrix contains information about no. observed and latent variables
    F <- mxFit@matrices$F@values 
    ny <- nrow(F) 
    ne <- ncol(F)

    which.s <- which(vech(mxFit@matrices$S@free))
    which.a <- which(mxFit@matrices$A@free)

    # The order in N&S is: [ vec LY, vech PH, vec BE, vech PS ]
    # In the RAM model, LY is the filter matrix and loadings are parameterized
    #    in the BE.0 matrix (see above), while S = PH and PS = 0.

    # add no. indices in LY to which.s (i.e. PH)
    which.s <- (ny * ne) + which.s

    # add no. indices in LY and vech PH
    which.a <- (ny * ne) + ((ne * (ne + 1))/2) + which.a 

    # Return them in the order in which they are in the 
    #   parameter vector of OpenMx
    c(which.a, which.s) 
}
