

reivreg <- function( Mxx, Mxy, Suu, n, d, sv=0.0132, var.names = names(Mxx) ) {

# Clean user input: 
Mxy <- as.matrix( Mxy )
if( sum(dim(Suu) != dim(Mxx)) > 0 ) { Suu <- diag(Suu) }

# validate user input:
err<-c()
if( dim(Mxy)[1] != dim(Mxx)[1] ) { err <- c(err,"Covariance matrix of X and Y does not match covariance matrix of X. Please check input.") }
if( sum(dim(Suu) != dim(Mxx)) > 0 ) { err <- c(err,"Dimensions of covariance matrix of X and error covariance matrix do not match. Make sure Suu has the same dimensions (e.g. is diagonal for uncorrelated errors).") }
if( dim(Mxx)[1] != length(d) ) {err <- c(err,"There should be a degree of freedom for each diagonal element of Suu.")}
if( n<=0 ) {err<-c(err,"Number of observations can't be negative or zero.")}
if( sum(d<0)>0 ) {err<-c(err,"Degrees of freedom can't be negative.")}

if ( length(err)!= 0 ) {
 print(err)
 return(-1)
}


Mxx.orig <-  Mxx

# Choose a small sample correction factor a
#   Should be between K + 1 and K + 4 + d/n
#   Here K + 1 is used to be the same as Warren, White & Fuller 1974
a   <-  dim(Mxx)[1] + 1

# Insert eigenvalue calculation here
A.star  <- 1/(sqrt(as.vector(Suu)[Suu!= 0]))

# The covariance adjustment is performed,
#       Note the small sample adjustment 1-a/n
Mxx <- Mxx-( 1 - a/n )*Suu  

# The estimate of the linear regression coefficients are the same as 
#       the OLS one pretending that Mxx is the actual moment matrix:
b             <- solve(Mxx) %*% Mxy

# The OLS estimate without correction for measurement error would have been
b.OLS         <- round((solve(Mxx.orig) %*% Mxy),3)
# With standard errors:
se.OLS        <- round(sqrt(diag((1/n) * solve(Mxx.orig) * sv)),3)

# Standard errors calculation for the corrected estimates

## The residual sum of squares of the regression of Y on X is needed for 
##       the calculation of the standard errors:
#  sv <- 0.0132

## R is a thing depending on the degrees of freedom of Suu. Here the matrix R/n is calculated:
Rn            <- diag(as.vector( ( ( diag(Suu^2) * b^2 ) / d )))

## The variance matrix of the estimates is decomposed into three parts:

### The extra variance due to having estimated the measurement errors
due.to.est.errors <- 2 * solve(Mxx) %*% Rn %*% solve(Mxx)

### The extra variance due to measurement errors
due.to.errors     <-
        (1/n) * solve(Mxx) %*% ( (Suu*sv) + (Suu %*% b%*%t(b) %*% Suu)) %*% solve(Mxx)

### The regular variance of the estimates
due.to.est        <- (1/n) * solve(Mxx) * sv

### The total variance of the estimates is the sum of the above components:
Vb      <- due.to.est + due.to.errors + due.to.est.errors

### The square roots of the diagonal elements are the standard errors of the estimates:
se <- sqrt(diag(Vb))

# Print the results and the variance decomposition

est <- round( cbind(b,se), 3 ) 

decomp<-cbind( diag(due.to.est/Vb), diag(due.to.errors/Vb), diag(due.to.est.errors/Vb))

results<- as.data.frame( cbind( b.OLS,se.OLS,est, round(diag(Vb),3),100*round(decomp,3) ) )

names(results)<- c("b.naive","se b.naive","b","se b","var b","% sampling","% meas. errors","% sampling meas. errs")
row.names(results) <- var.names

return(list(b,se,results))
}
