#
#
#  Satorra-Bentler robust standard errors (and variance-covariance matrix
#  of the parameters) for openMx.
#
#  Author: Daniel Oberski
#  Date: 1 April 2010
#  License: same as R (GPL version 2): http://www.r-project.org/COPYING

# ML covariance matrix
cov.ml <- function(X, ...) {
    n <- NROW(X)
    cov(X, ...) * (n - 1) / n
}

# Weighted sample mean. 
mean.wt <- function(x, wt=NULL) {
    if (is.null(wt)) wt <- rep(1, length(x))
    sum(wt * x) / sum(wt)
}

# Normal-theory Gamma matrix
#   e.g. Satorra and Bentler 1990: 239, eq. 2.10
get.gamma.star <- function(S, vec.or.vech='vec') {
  gam <- Matrix(2 * S %x% S)
  if(vec.or.vech == 'vech') {
    D.plus <- mp.inv(duplication.matrix(ncol(S)))
    gam <- D.plus %*% gam %*% t(D.plus)
  }
  gam
}

# Inverse of NT Gamma matrix is more easily obtained
#   via inverse of covariance matrix:
get.gamma.star.inv <- function(S, vec.or.vech='vec') {
  Si <- solve(S)
  Gami <- (Si %x% Si) / 2
  if(vec.or.vech == 'vech') {
    D <- duplication.matrix(ncol(S))
    Gami <- t(D) %*% Gami %*% D
  }
  Gami
}


# This function calculates the "asymptotic covariance matrix". 
#   This is the covariance matrix of the sample variances and covariances.
# An alternative to this matrix is to use the hessian from a polycor::hetcor
#  output, or to use the Normal-Theory gamma matrix of the original covariances
# in the case of covariance reduction. 
# This function provides a consistent estimate of avar(S) under non-normality
#   see, for example, Satorra and Bentler 1990: 238, eq. 2.4 
#   or Fuller 1987: 332 
#   (He uses the unbiased version, while we use the maximum likelihood version, 
#        the difference being a division by n^2 or by (n - 2)(n - 1) )
get.gamma <- function(X.raw, wt=NULL, matconstructor=Matrix) {
  p <- NCOL(X.raw)
  pstar <- p * (p + 1)/2
  # If no weights are given, weights are set to 1 for all cases:
  if (is.null(wt)) wt <- rep(1, NROW(X))
  sum.wt <- sum(wt) # will equal sample size if no weights

  Dplus <- mp.inv(duplication.matrix(p))

  # Vector of (weighted) means of X
  mean.X <- apply(X.raw, 2, function(x) mean.wt(x, wt))
  # Z is X expressed in deviations from (weighted) mean
  Z <- sweep(X.raw, 2, mean.X)

  B <- apply(Z, 1, function(Zi) Dplus %*% cbind(as.numeric(Zi %*% t(Zi))))

  # Weight B
  B <- t(apply(B, 1, function(Bi) wt * Bi))

  # This is Muthen & Satorra eq. 18 but without the aggregation
  # Without weights it would be rowMeans(B)
  Bbar <- rowSums(B) / sum.wt
  
  Gvec <- rowSums(
            apply(B, 2, 
                function(bt) cbind(bt - Bbar) %*% rbind(bt - Bbar) 
  )) / sum.wt
  
  matconstructor(Gvec, pstar, pstar)
}

get.Bbar <- function(Z, clusters) {
    J <- length(unique(clusters)) # number of clusters
    mean.X <- apply(Z, 2, mean)
    Dplus <- mp.inv(duplication.matrix(NCOL(Z)))

    Bsum <- by(Z, clusters, function(Zj) {
        # Z is X expressed in deviations from overall (weighted) mean
        Zj <- sweep(Zj, 2, mean.X)
        B <- apply(Zj, 1, function(Zi) Dplus %*% cbind(as.numeric(Zi %*% t(Zi))))
        B <- t(apply(B, 1, function(Bi) Bi))
        rowSums(B)/NROW(Zj)
    })

    Reduce("+", Bsum) / J
}

# Muthen & Satorra (p. 283) clustered gamma matrix
# Sampling weights can also be incorporated via argument wt.
get.gamma.clustered <- function(X.raw, clusters, 
        wt=NULL, normal.theory=FALSE, PS=NULL) {
    J <- length(unique(clusters)) # number of clusters
    p <- NCOL(X.raw) # number of observed variables
    pstar <- p * (p + 1)/2 # number of unique cov matrix elements
    n <- sum.wt <- NROW(X.raw)

    Dplus <- mp.inv(duplication.matrix(p))
    Z <- X.raw
    # Vector of overall means of X
    mean.X <- apply(X.raw, 2, mean)

    Bbar <- get.Bbar(Z, clusters)

    G <- by(Z, clusters, function(Zj) {
       if(normal.theory) { # Warning: no weights!
           S <- cov.ml(Zj)
           if (!is.null(PS)) S <- S - PS # if meas error is given subtract it
           return(nrow(Zj) * get.gamma.star(S, 'vech'))
       }
       else {# Non-normal data
          # Z is X expressed in deviations from overall (weighted) mean
          Zj <- sweep(Zj, 2, mean.X)
          B <- apply(Zj, 1, function(Zi) Dplus %*% cbind(as.numeric(Zi %*% t(Zi))))
          B <- t(apply(B, 1, function(Bi) Bi))

          # This should be overall Bbar not within
        #Bbar <- rowSums(B) / sum.wt

          Gvec <- rowSums(
                    apply(B, 2, 
                        function(bt) cbind(bt - Bbar) %*% rbind(bt - Bbar) 
          )) / sum.wt
          
          matrix(Gvec, pstar, pstar)
       }
    })

    (J * Reduce("+", G))/ (J - 1)
}




# Get the Satorra-Bentler variance matrix of the
#  parameter estimates from an openmx fit object.
# This provides robust standard errors.
get.mx.robust.variance <- function(mxFit, Gamma, use.expected=TRUE) {    
  n <- mxFit@data@numObs
  if (use.expected) {
    Sigma <- Matrix(mxExpected(mxFit))
  }
  else {
    if(mxFit@data@type != 'cov')
      stop("Use observed only works with covariance type models")
    Sigma <- Matrix(mxFit@data@observed)
  }
  
  # If the cov-ML objective criterion was used, the weight matrix equals:
  #W <- Matrix(2 * Sigma %x% Sigma)
  #W.i <- solve(W)
  # Stabler inversion of W via inversion of Sigma (Satorra & Bentler 1990: 239):
  Sigma.inv <- solve(Sigma) 
  W.i <- Matrix((Sigma.inv %x% Sigma.inv)/2)

  # Obtain derivatives of implied covariances wrt free parameters
  Delta <- get.Delta.analytical(mxFit)

  # Satorra & Bentler 1990: 239, eq. 2.12a
  Bread <- solve(t(Delta) %*% W.i %*% Delta)
  Filling <- t(Delta) %*% W.i %*% Gamma %*% W.i %*% Delta

  V.robust <- (1/n) * Bread %*% Filling %*% Bread

  V.robust
}


dvec.Sigma.LISREL <- function(LY, PH, BE, PS,
                              output.method=cBind,
                              vech.Sigma=FALSE,
                              vech.PH=TRUE) {
  # Note: BE is already (I - BE.0)
  Bi <- solve(BE)

  LBi <- (LY %*% Bi)
  p <- nrow(LY) # no. observed variables
  q <- ncol(LY) # no. latent variables
  I.pp <- Id(p^2)
  Kplus <- I.pp + Kmat(p, p)

  if(vech.Sigma)
    L <- Matrix(elimination.matrix(p), sparse=TRUE)
  if(vech.PH) {
    Dp <- Matrix(duplication.matrix(p), sparse=TRUE)
    Dq <- Matrix(duplication.matrix(q), sparse=TRUE)
  }
  
  # Derivatives wrt vec PH:
  dvec.PH <-  LBi %x% LBi
  if(vech.PH)
    dvec.PH <- dvec.PH %*% Dq
  if(vech.Sigma)
    dvec.PH <- L %*% dvec.PH
  
  # Derivatives wrt vec BE
  R <- (LBi %*% PH %*% t(Bi))
  dvec.BE <- Kplus %*% (R %x% LBi)
  if(vech.Sigma)
    dvec.BE <- L %*% dvec.BE
  
  # Derivatives wrt vec LY
  dvec.LY <- Kplus %*% (R %x% Id(p))
  if(vech.Sigma)
    dvec.LY <- L %*% dvec.LY
  
  # Derivatives wrt vec PS
  dvec.PS <- I.pp
  if(vech.PH)
    dvec.PS <- dvec.PS %*% Dp
  if(vech.Sigma)
    dvec.PS <- L %*% dvec.PS

  output.method('dvec.LY'=dvec.LY, 'dvec.PH'=dvec.PH,
        'dvec.BE'=as.matrix(dvec.BE),'dvec.PS'=dvec.PS)
}

get.Sigma.LISREL <- function(LY, PH, BE, PS) {
  Bi <- solve(BE)
  LY %*% Bi %*% PH %*% t(Bi) %*% t(LY) + PS
}

get.gradient.LISREL <- function(LY, PH, BE, PS, S, W=NULL) {
  
  Sigma <- get.Sigma.LISREL(LY, PH, BE, PS)
  
  if (is.null(W))
    W <- get.gamma.star.inv(Sigma)

  # take the vech of PH and PS so that the gradient expression
  # is the same for all parameters
  G.list <- dvec.Sigma.LISREL(LY, PH, BE, PS, list, FALSE, TRUE) 

  z <- vec(S) - vec(Sigma)

  zW <- t(z) %*% W

  grad <- lapply(G.list, function(G) -1 * zW %*% G)
  grad
}

get.num.params <- function(LY, PH, BE, PS) {
  q <- ncol(PH)
  p <- ncol(PS)
  length(LY) + length(BE) + q*(q+1)/2 + p*(p+1)/2
}


# Variance of parameters of a LISREL model under asymptotic
#   optimality with inverse of 4th order moments Gi
V.AO <- function(LY, PH, BE.0, PS, Gi, n, free.inds=NULL) {
  BE <- Id(ncol(BE.0)) - BE.0 # cf. Neu & Sat p. 58
  Bi <- solve(BE)

  if (is.null(free.inds)) { # take all, 1:q
    free.inds <- 1:prod(c(dim(LY), dim(PH), dim(BE.0), dim(PS)))
  }  
  D <- dvec.Sigma.LISREL(LY, PH, BE, PS, free.inds)
    
  avar.theta <- solve(t(D) %*% Gi %*% D) / n

  avar.theta
}


# Neudecker & Satorra 1989:59, eq. 3
# note: we use vec Sigma rather than vech Sigma so the
#       elimination matrices L in eq. 3 and 4 dissapear.
get.Delta.analytical <- function(mxFit, which.thetas="free") {
  # Give analytical derivatives wrt vec A and vech S .
  #   F is skipped since it is not parameterized in RAM.
  # Note that A is called BETA in LISREL and S is PHI
  which.thetas <- tolower(which.thetas)
  if (!which.thetas %in% c('free', 'fixed', 'all'))
    stop('which.thetas should be one of "free", "fixed" or "all".')
  
  A.0 <- Matrix(mxEval(A, mxFit))
  A <- Id(ncol(A.0)) - A.0 # cf. Neu & Sat p. 58
  Ai <- solve(A)
  S <- Matrix(mxEval(S, mxFit))
  F <- Matrix(mxEval(F, mxFit))

  # Figure out whether to take derivatives wrt free parameters,
  #   fixed parameters, or both
  if(which.thetas == 'free') {
    which.s <- as.numeric(mxFit@matrices$S@free)
    which.a <- as.numeric(mxFit@matrices$A@free)
  }
  else if (which.thetas == 'fixed') {
    which.s <- as.numeric(!mxFit@matrices$S@free)
    which.a <- as.numeric(!mxFit@matrices$A@free)
  }
  else {
    which.s <- rep(1, length(S))
    which.a <- rep(1, length(A))
  }
 
  FAi <- (F %*% Ai)
  
  # Derivatives wrt vec S:
  Ds <-  FAi %x% FAi
    
  # Derivatives wrt vec A
  p <- nrow(F) # no. observed variables
  R <- (FAi %*% S %*% t(Ai))  %x%  FAi
  Kplus <- Id(p^2) + Kmat(p, p)
  Da <- (Kplus %*% R)
  
  # Select only relevant columns of S and A:
  Ds <- Ds[, as.logical(which.s)]  # checks out against numerical derivs.
  Da <- Da[, as.logical(which.a)]  # checks out " " "
  
  # G is a partitioned matrix (Da, Ds)
  G <- cBind(Da, Ds)

  G
}
