# Demonstration of covariance reduction method of correction for measurement 
# error. After the correction a correction to the standard errors is necessary
# to provide measurement error-corrected standard errors. 
# These corrected standard errors still do not take into account that the 
# measurement error variance used to correct was estimated, not known. 
# In this file a bootstrap procedure is used to estimate the total standard 
# error including uncertainty due to the estimation of the measurement errors.
# For more information as well as the rationale behind the procedures, see

# Oberski, D. & A. Satorra (forthcoming). 
#  How to take the variability of measurement error variances into account.
# http://daob.org/ 

# Author: Daniel Oberski
# Date: 19 April 2010
# License: GPL >= 2 

library(OpenMxExtras)

data(guillen)

# Names of the latent and observed variables in the model
latents <- c("Achievement", "Power")
manifests <- colnames(guillen)

# Asymptotic covariance matrix under multivariate normality:
# Note that the original not the reduced covariance matrix is used!
Gamma <- get.gamma.star(guillen)

# Reliability of SCF as given in the paper
rel.scf <- 0.76

# Reduce the variance of SCF to its true variance
guillen.reduced <- guillen
guillen.reduced[1] <- rel.scf * guillen[1] 
  
# Define the OpenMx model with SCF as an observed variable:
guillenModel <- 
  mxModel(name="Guillen",
          type="RAM",
          manifestVars = manifests,
          latentVars = latents,
          mxPath(from="Achievement", to="INI", free=FALSE, values=1),
          mxPath(from="Achievement",
                 to=c("OPT", "ACH", "CHG")),
          mxPath(from="Power", to="DEV", free=FALSE, values=1),
          mxPath(from="Power",
                 to=c("CONF", "INF", "LEAD")),
          mxPath(from="SCF", to=latents),
          mxPath(from=manifests, arrows=2),
          mxPath(from=latents, arrows=2),
          mxData(observed=round(guillen.reduced,6), type="cov", numObs=223)
          )

# Run the model using OpenMx
guillenFit <- mxRun(guillenModel)
summary(guillenFit)

# Parameter estimates
theta.estimate <- guillenFit@output$estimate

# "Naive" standard errors (not taking into account measurement error 
#  correction or the estimation of the error
sd.naive <- guillenFit@output$standardErrors

# Variance-covariance matrix of the parameter estimates 
#    taking into account the covariance reduction used
V.corrected <- get.mx.robust.variance(guillenFit, Gamma)

# Standard errors corrected for the covariance reduction
sd.corrected <- sqrt(diag(V.corrected))


# Show differences
print(cbind(estimates, sd.naive, sd.corrected))


# Define a function to run the above analysis
estimate.model <- function(rel.scf, standard.errors=FALSE) {

  guillen.reduced <- guillen
  guillen.reduced[1] <- rel.scf * guillen[1] # reduce var of SCF
  
  guillenModel <- mxModel(name="Guillen",
                          type="RAM",
                          manifestVars = manifests,
                          latentVars = latents,
                          mxPath(from="Achievement", to="INI", free=FALSE, 
                              values=1),
                          mxPath(from="Achievement",
                                 to=c("OPT", "ACH", "CHG")),
                          mxPath(from="Power", to="DEV", free=FALSE, values=1),
                          mxPath(from="Power",
                                 to=c("CONF", "INF", "LEAD")),
                          mxPath(from="SCF", to=latents),
                          mxPath(from=manifests, arrows=2),
                          mxPath(from=latents, arrows=2),
                          mxData(observed=round(guillen.reduced,4), 
                              type="cov", numObs=223)
                          )
  if (!standard.errors) {
      guillenModel <- mxOption(guillenModel, "Standard Errors", "No")
      guillenModel <- mxOption(guillenModel, "Calculate Hessian", "No")
  }
  guillenFit <- mxRun(guillenModel)
  estimates <- guillenFit@output$estimate
  
  V.analytical <- get.mx.robust.variance(guillenFit, Gamma)
  sd.correct <- sqrt(diag(V.analytical))
  
  list(est=estimates, sd.correct=sd.correct)
}

# The time this takes * nsim is approximately how long the bootstrap will
#  take. This is 2 minutes on my machine with 200 replications.
system.time(main.model <- estimate.model(rel.scf))

# Run bootstrap on the model
nsim <- 500  # number of bootstrap simulations
# The standard error of the reliability estimate of the variable SCF:
sd.rel <- 0.04 
# The average (estimate) of the reliability of the variable SCF:
mu.rel <- 0.76
# Vector of simulated reliabilities:
rels.sim <- rnorm(nsim, mu.rel, sd.rel)

p <- length(theta.estimate) # Number of parameter in the model

# Run the bootstrap. This takes a very long time!
system.time(res <- mapply(function(rel.sim) {
  # TryCatch is used because some problems may not converge
  tryCatch(return(estimate.model(rel.sim)), error = function(e)
             # If no convergence add NA's to the result data frame
             list(est=rep(NA, p), sd.naive=rep(NA, p), sd.correct=rep(NA, p))
           )
}, rels.sim))

# Bootstrap estimates
estimates.boot <- t(matrix(unlist(res[1,]), p, nsim))    
# Bootstrap naive standard errors
sd.naive.boot <- t(matrix(unlist(res[2,]), p, nsim))     
# Bootstrap measurement error-corrected standard errors
#sd.correct.boot <- t(matrix(unlist(res[3,]), p, nsim))   


# The extra variance due to the estimation of the reliability is estimated
#   consistently by the bootstrap variance over replications of the estimates
extra.var <- apply(estimates.boot, 2, var)
total.var <- sd.corrected^2  + extra.var

# Bootstrap estimates are on average the same even when the measurement error is
#    varied.
print(
    data.frame(Model.estimates=theta.estimate, 
      Bootstrap.estimates=apply(estimates.boot, 2, mean))
)

# Comparison between the naive, measurement error-corrected, and total standard 
#   errors
results <- 
 data.frame(Estimate=round(theta.estimate,3), 
   # Does not take into account the correction for measurement error:
   sd.naive=round(sd.naive,3),        
   # Does not take into account the measurement error was estimated not known:
   sd.measerr=round(sd.corrected,3),  
   # Variance comprising sampling, measurement error correction, 
   #       and unknown (estimated) measurement error
   sd.total=round(sqrt(total.var),3),
   sampling = round(100*sd.naive^2/total.var),
   meas.error = round(100*(sd.corrected^2 - sd.naive^2)/total.var),
   est.error = round(100*extra.var/total.var)
   )

print(results)
