# Demonstration of fixed single indicators method of correction for measurement 
# error. 
# The standard errors still do not take into account that the 
# measurement error variance used to correct was estimated, not known. 
# In this file a bootstrap procedure is used to estimate the total standard 
# error including uncertainty due to the estimation of the measurement errors.
# For more information as well as the rationale behind the procedures, see

# Oberski, D. & A. Satorra (forthcoming). 
#  How to take the variability of measurement error variances into account.
# http://daob.org/ 

# Author: Daniel Oberski
# Date: 21 April 2010
# License: GPL >= 2 

library(snowfall)

# initialize multicore; leave one for other work
sfInit(parallel = TRUE, cpus = 7)


# defines estimate.model, errorfunc, variance.scf, guillen, nobs, latents,
#    manifests, p, rel.scf, sd.rel.scf;
# imports OpenMxExtras
source('init.R')

# Variance of the variance under normality (e.g. Yuan & Hayashi 2007)
var.variance.scf <- (variance.scf^2 * 2) / (nobs - 1)

# Asymptotic covariance matrix under multivariate normality:
Gamma <- get.gamma.star(guillen)

# Reliability of SCF as given in the paper
rel.scf <- 0.76

# Standard error of the reliability (calculated from Guillen et al.)
sd.rel.scf <- 0.04

# Error variance equals (see for example Saris & Gallhofer 2007): 
err.var.scf <- (1 - rel.scf) * variance.scf

stopifnot(abs(err.var.scf - 0.8664) < 1e-4) # Should equal .87

# Variance of the error variance of SCF is approx. 0.03
var.err.var.scf <- cbind(get.err.var(variance.scf, nobs, rel.scf, sd.rel.scf^2)[2])


# Define the OpenMx model with SCF as an observed single indicator
# of the factor Self-confidence:
guillenModel <- 
  mxModel(name="Guillen",
          type="RAM",
          manifestVars = manifests,
          latentVars = latents,
          mxPath(from="Achievement", to="INI", free=FALSE, values=1),
          mxPath(from="Achievement",
                 to=c("OPT", "ACH", "CHG")),
          mxPath(from="Power", to="DEV", free=FALSE, values=1),
          mxPath(from="Power",
                 to=c("CONF", "INF", "LEAD")),
          mxPath(from="Self-confidence", to=latents[1:2]),
          mxPath(from="Self-confidence", to="SCF", free=FALSE, values=1),
          mxPath(from=manifests, arrows=2),
          mxPath(from="SCF", arrows=2, free=FALSE, values=err.var.scf),
          mxPath(from=latents, arrows=2),
          mxData(observed=round(guillen, 6), type="cov", numObs=223)
          )

# Run the model using OpenMx
guillenFit <- mxRun(guillenModel)

# Parameter estimates
theta.estimate <- guillenFit@output$estimate

# "Naive" standard errors (not taking into account the estimation of the error)
# Contrary to the naive sd's in the covariance reduction example, these sd's 
# do already take the correction for measurement error into account.
sd.naive <- guillenFit@output$standardErrors

# The time this takes * nsim is approximately how long the bootstrap will
#  take. This is 2 minutes on my machine with 200 replications.
system.time(main.model <- estimate.model(rel.scf)) # defined in init.R

# Run bootstrap on the model
nsim <- 1000  # number of bootstrap simulations
# The standard error of the reliability estimate of the variable SCF:
sd.rel <- 0.04 
# The average (estimate) of the reliability of the variable SCF:
mu.rel <- 0.76
# Vector of simulated reliabilities:
rels.sim <- rnorm(nsim, mu.rel, sd.rel)

# Run the bootstrap. This takes a very long time!
system.time(res <- sfSapply(rels.sim, function(rel.sim) {
  # define variables and load libraries
  source('~/work/meas_error/R/OpenMxExtras/demo/init.R')  
  # TryCatch is used because some problems may not converge
  tryCatch(return(estimate.model(rel.sim)), error = errfunc)
}))

# Bootstrap estimates
estimates.boot <- t(matrix(unlist(res[1,]), p, nsim))

# Bootstrap naive standard errors
sd.naive.boot <- t(matrix(unlist(res[2,]), p, nsim))

extra.var <- apply(estimates.boot, 2, var)
total.var <- sd.naive^2  + extra.var

# Bootstrap estimates are on average the same even when the measurement error is
#    varied.
print(
    data.frame(Model.estimates=theta.estimate, 
      Bootstrap.estimates=apply(estimates.boot, 2, mean))
)

# Comparison between the naive and total standard errors
results <- OpenMx:::parameterList(guillenFit, FALSE)
results$sd.corrected <- sqrt(total.var)
results[, 5:7] <- round(results[, 5:7], 3)
results$est.error <- round(100*extra.var/total.var)

results <- results[, -c(1,2)]
print(results)

sfStop()
