# Monte Carlo experiment

library(OpenMxExtras)
library(mvtnorm)


reliability <- c('High'     = 0.9, 
                 'Middle'   = 0.7, 
                 'Low'      = 0.4)

standard.error <- c('Large'         = 0.15 , 
                    'Medium'        = 0.05 , 
                    'Small'         = 0.01 ,
                    'Very small'    = 0.001)

sample.size <- c('Large'     = 1000, 
                 'Medium'    =  500, 
                 'Small'     =   50)


trials <- expand.grid(reliability=reliability, 
                      standard.error=standard.error, 
                      sample.size=sample.size)

num.simulations <- 10

effect.size <- 0.5

manifests <- paste('y', 1:5, sep='')
latents <- paste('n', 1:5, sep='')


get.population.covariance <- function(effect.size, reliability) {
    stopifnot(abs(effect.size - .5) < 1e-6) # Quick fix to prevent mistakes
    BE.0 <- Matrix(c(0, rep(effect.size, 4), 
                   0,0,0, effect.size, 0, 
                   0,0,0,0, effect.size, 
                   rep(0,10)), 5, 5)
    BE <- Id(5) - BE.0
    Bi <- solve(BE)

    PH <- Matrix(diag(c(1, 1 - effect.size^2, 1 - effect.size^2, 
                effect.size^2, effect.size^2))) # wrong but works for .5

    LY <- Matrix(diag(sqrt(reliability), 5))
    PS <- Matrix(diag(1 - reliability, 5))

    S <- LY %*% Bi %*% PH %*% t(Bi) %*% t(LY) + PS
    colnames(S) <- rownames(S) <- manifests

    return(as.matrix(S))
}

get.model <- function(reliability, S, sample.size, itnum) {
    errvar <- (1 - reliability) * diag(S)

    simModel <- 
      mxModel(name=paste("Simulation", itnum, sep=""),
              type="RAM",
              manifestVars = manifests,
              latentVars = latents,
              mxPath(from="n1", to=c("n2", "n3", 'n4', 'n5')),
              mxPath(from="n2", to="n4"),
              mxPath(from="n3", to="n5"),
              mxPath(from="y1", arrows=2, free=FALSE, values=errvar[1]),
              mxPath(from="y2", arrows=2, free=FALSE, values=errvar[2]),
              mxPath(from="y3", arrows=2, free=FALSE, values=errvar[3]),
              mxPath(from="y4", arrows=2, free=FALSE, values=errvar[4]),
              mxPath(from="y5", arrows=2, free=FALSE, values=errvar[5]),
              mxPath(from=latents, to=manifests, free=FALSE, values=1),
              mxPath(from=latents, arrows=2),
              mxData(observed=round(as.matrix(S),6), 
                  type="cov", numObs=sample.size)
    )

    simModel 
}

sample.covmat <- function(population.covmat, sample.size) {
    # Draw a random sample from a mv normal distribution and return the covariance
    Y <- rmvnorm(sample.size, rep(0, 5), population.covmat)
    S <- cov(Y)
    colnames(S) <- rownames(S) <- manifests

    return(S)
}

run.iteration <- function(trial, itnum) {
    
    # Draw a random reliability value
    sample.reliability <- rnorm(1, trial[['reliability']], trial[['standard.error']])

    # Calculate population covariance matrix given the sample reliability
    Sigma <- get.population.covariance(effect.size, sample.reliability)

    # Draw a random sample observed matrix from MVNORM with population covmatrix
    S <- sample.covmat(Sigma, trial[['sample.size']])

    # Estimate the model using OpenMx
    fit <- mxRun(get.model(trial[['reliability']], S, trial[['sample.size']], itnum))

    # Determine parameter indices of theta.eps
    fixed.params <-   c(get.par.index('S', 1, 1, fit), 
                        get.par.index('S', 2, 2, fit), 
                        get.par.index('S', 3, 3, fit), 
                        get.par.index('S', 4, 4, fit), 
                        get.par.index('S', 5, 5, fit))
    
    # Obtain the estimated variance of the error variance in this sample
    var.err.var <- diag(get.err.var(S[1], trial[['sample.size']], 
            trial[['reliability']], trial[['standard.error']]^2)[2], 5)

    # Calculate uncorrected and corrected analytic standard errors
    V <- get.var.corrected(mxFit=fit, fixed.params=fixed.params, 
        V.fixed=cbind(var.err.var))

    list('estimate'=fit@output$estimate,
         'sd.naive.numeric'=as.vector(fit@output$standardErrors),
         'sd.naive.analytic'=sqrt(diag(V$uncorrected)),
         'sd.corrected.analytic'=sqrt(diag(V$corrected)))
}

run.simulation <- function(trial) {
    mapply(function(itnum) run.iteration(trial, itnum), 1:num.simulations)
}

analyze.trial.result <- function(trial.result, ndigits=3) {
    estimates <- matrix(unlist(trial.result[1, ]),
            num.simulations, 11, byrow=TRUE)[, 1:6]
    se.naive.numeric <- matrix(unlist(trial.result[2, ]),
            num.simulations, 11, byrow=TRUE)[, 1:6]
    se.naive.analytic <- matrix(unlist(trial.result[3, ]),
            num.simulations, 11, byrow=TRUE)[, 1:6]
    se.corrected.analytic <- matrix(unlist(trial.result[4, ]),
            num.simulations, 11, byrow=TRUE)[, 1:6]

    simulation.ses <- data.frame(
        simulation.sd = apply(estimates, 2, sd),
        average.naive.se.numeric = apply(se.naive.numeric, 2, mean),
        average.naive.se.analytic = apply(se.naive.analytic, 2, mean),
        average.corrected.se = apply(se.corrected.analytic, 2, mean)
    )

    print(round(simulation.ses, ndigits))
}

# Run the simulation
#result <- apply(trials, 1, run.simulation)
#lapply(results, analyze.trial.result)
