\name{get.err.var}
\alias{get.err.var}
\title{Calculate error variance and its standard error}
\description{
    Given a reliability ("reliability ratio") and an observed variance,
    calculate the value of the error variance and its variance (square 
    of the standard error).
}
\details{
    Structural equation models are formulated in terms of a covariance
    structure. Thus in order to perform an analysis that takes into account
    the reliability of a measure, that reliability, which is a standardized
    quantity, must be transformed into an unstandardized error variance. 

    One then creates a model with a latent variable for the error-prone 
    measure and specifies the measure as a single indicator of that latent
    variable. The loading is fixed to unity (the number 1) and the error 
    variance to the number returned by this function.

    The reliability and observed variance are sample quantities and therefore
    contain errors as well. This function will propagate these uncertainties
    into the standard error of the error variance. This value can then be used
    to correct the standard errors of the model parameters.
}
\usage{
    get.err.var(obs.var, nobs, rel, var.rel, excess.kurtosis = 0)
}
\arguments{
\item{obs.var}{Observed variance of the measure, for example as obtained by 
    var(x).}
\item{nobs}{Number of sample observations used to estimate the 
    observed variance of the measure.}
\item{rel}{Reliability of the measure. This equals var(T)/var(x), where T is
    the so-called "true score". In the terminology of Saris and Gallhofer 2007, 
    the "quality" of the measure. Note: this quality equals the square of 
    the "quality coefficient".}
\item{var.rel}{Variance of the reliability estimate.}
\item{excess.kurtosis}{Optionally for non-normally distributed data, 
    one can provide the excess kurtosis of the observed data used to estimate
        the sample variance. Kurtoses larger than 0 will increase the 
        standard error.}
}
\value{
    \item{error.variance}{The calculated value of the error variance estimate.}
    \item{var.error.variance}{The variance (square of the standard error) of the
        error variance estimate.}
}
\seealso{
\code{\link{get.stderr.corrected}}, \code{\link{guillen}}.
}
\examples{
    data(guillen)
    # Observed variance equals 3.61
    (obs.variance <- guillen[1])

    # Error variance for the SCF variable. The sample size (223),
    #   reliability (0.76), and standard error (0.04) of the reliability
    #   are taken directly or calculated indirectly from 
    # Guillen et al. (2009)
    get.err.var(obs.variance, 223, 0.76, 0.04^2)

    # If the SCF variable had a kurtosis of 6 (corresponding to a 
    #  t-distribution with 5 df), the error variance estimate
    #  would not change, but the variance of this estimate would increase:
    get.err.var(obs.variance, 223, 0.76, 0.04^2, excess.kurtosis = 6)
}
\keyword{get.err.var}
\references{
    Oberski, D. and A. Satorra (2010). How to take the variability of measurement
        error variances into account. http://daob.org/
}
\author{Daniel Oberski \email{daniel.oberski@upf.edu}}
