/*
    Commutation matrix

    Author: Daniel Oberski
    Date: 4 April 2010
    Original algorithm: Frederick Novomestky
    License: GPL 2
    

    Outputs 1-based indices indicating which elements of the otherwise 0
    m*n * m*n matrix K in R should be set to 1. In combination with a
    sparse Matrix from the Matrix package this provides a fast way and
    memory-efficient way of creating a commutation matrix K.
    
    Reference:

    Magnus & Neudecker (1979). The commutation matrix: some properties
        and applications. The annals of statistics, vol. 7, no. 2 pp. 381-394.
*/
void commutation_indices(const int *m, const int *n, double *Ki) {
    int i, j, p; 
    int r = 0,  c = 0;
    p = (*m) * (*n);

    for(i=0; i<*m; i++) {
        c = i;
        for(j=0; j<*n; j++) {
            Ki[j*(*m) + i] = c*p + r + 1; // R uses 1-based index
            r++;
            c += *m;
        }
    }
}

/* Literal C-transscription of the commutation.matrix function found 
    int the matrxicalc package. Funnily enough this C function, 
    when called from R, is _slower_ than the pure R version. 
    This is mostly due to the need for copying the large matrix K
    back and forth. For the same reason it also requires a lot of memory
    to operate.
*/
void commutation_matrix(const int *m, const int *n, double *K) {
    int i, j, p; 
    int r = 0,  c = 0;
    p = (*m) * (*n);

    for(i=0; i<*m; i++) {
        c = i;
        for(j=0; j<*n; j++) {
            K[c*p + r] = 1;
            r++;
            c += *m;
        }
    }
}

