library(OpenMxExtras)

data(guillen)

# Names of the latent and observed variables in the model
latents <- c("Achievement", "Power")
manifests <- colnames(guillen)

# Asymptotic covariance matrix under multivariate normality:
# Note that the original not the reduced covariance matrix is used!
Gamma <- get.gamma.star(guillen)

# Reliability of SCF as given in the paper
rel.scf <- 0.76

# Reduce the variance of SCF to its true variance
guillen.reduced <- guillen
guillen.reduced[1] <- rel.scf * guillen[1] 
  
# Define the OpenMx model with SCF as an observed variable:
guillenModel <- 
  mxModel(name="Guillen",
          type="RAM",
          manifestVars = manifests,
          latentVars = latents,
          mxPath(from="Achievement", to="INI", free=FALSE, values=1),
          mxPath(from="Achievement",
                 to=c("OPT", "ACH", "CHG")),
          mxPath(from="Power", to="DEV", free=FALSE, values=1),
          mxPath(from="Power",
                 to=c("CONF", "INF", "LEAD")),
          mxPath(from="SCF", to=latents),
          mxPath(from=manifests, arrows=2),
          mxPath(from=latents, arrows=2),
          mxData(observed=round(guillen.reduced,6), type="cov", numObs=223)
          )

# Run the model using OpenMx
guillenFit <- mxRun(guillenModel)

# Parameter estimates
estimates <- guillenFit@output$estimate

# "Naive" standard errors (not taking into account measurement error 
#  correction or the estimation of the error
sd.naive <- guillenFit@output$standardErrors

# Variance-covariance matrix of the parameter estimates 
#    taking into account the covariance reduction used
V.corrected <- get.mx.robust.variance(guillenFit, Gamma)

# Standard errors corrected for the covariance reduction
sd.corrected <- sqrt(diag(V.corrected))

print(cbind(estimates, sd.naive, sd.corrected))

sd.corrected.should.be <-
c(0.06134, 0.05415, 0.06727, 0.04793, 0.05949, 0.16464, 0.17048, 
    0.22521, 0.34188, 0.11237, 0.12872, 0.04943, 0.09895, 0.12774, 
    0.25775, 0.15589, 0.23150, 0.12528, 0.05655)

stopifnot(abs(sd.corrected - sd.corrected.should.be) < 1e-5)
