library(OpenMxExtras)

data(guillen)
variance.scf <- guillen[1] # The observed variance of the variable SCF

# Sample size for the Guillen et al. data is 223:
nobs <- 223

# Variance of the variance under normality (e.g. Yuan & Hayashi 2007)
var.variance.scf <- (variance.scf^2 * 2) / (nobs - 1)

# Names of the latent and observed variables in the model
latents <- c("Achievement", "Power", "Self-confidence")
manifests <- colnames(guillen)

# Asymptotic covariance matrix under multivariate normality:
Gamma <- get.gamma.star(guillen)

# Reliability of SCF as given in the paper
rel.scf <- 0.76

# Standard error of the reliability (calculated from Guillen et al.)
sd.rel.scf <- 0.04

# Error variance equals (see for example Saris & Gallhofer 2007): 
err.var.scf <- (1 - rel.scf) * variance.scf

stopifnot(abs(err.var.scf - 0.8664) < 1e-4) # Should equal .87

# Variance of the error variance of SCF is approx. 0.03
var.err.var.scf <- cbind(get.err.var(variance.scf, nobs, rel.scf, sd.rel.scf^2)[2])


# Define the OpenMx model with SCF as an observed single indicator
# of the factor Self-confidence:
guillenModel <- 
  mxModel(name="Guillen",
          type="RAM",
          manifestVars = manifests,
          latentVars = latents,
          mxPath(from="Achievement", to="INI", free=FALSE, values=1),
          mxPath(from="Achievement",
                 to=c("OPT", "ACH", "CHG")),
          mxPath(from="Power", to="DEV", free=FALSE, values=1),
          mxPath(from="Power",
                 to=c("CONF", "INF", "LEAD")),
          mxPath(from="Self-confidence", to=latents[1:2]),
          mxPath(from="Self-confidence", to="SCF", free=FALSE, values=1),
          mxPath(from=manifests, arrows=2),
          mxPath(from="SCF", arrows=2, free=FALSE, values=err.var.scf),
          mxPath(from=latents, arrows=2),
          mxData(observed=round(guillen, 6), type="cov", numObs=223)
          )

# Run the model using OpenMx
guillenFit <- mxRun(guillenModel)

# Parameter estimates
estimates <- guillenFit@output$estimate

# "Naive" standard errors (not taking into account the estimation of the error)
# Contrary to the naive sd's in the covariance reduction example, these sd's 
# do already take the correction for measurement error into account.
sd.naive <- guillenFit@output$standardErrors

# Element 1,1 of the S matrix is the error variance of SCF and has been fixed:
fixed.params <- get.par.index('S', 1, 1, guillenFit)

# (this should yield 1 + 12*9 = 109)
stopifnot(fixed.params == c(109))

# Variance-covariance matrix of the parameter estimates 
#    taking into account that the measurement error was estimated not known
# Takes 1 second on my machine:
V.total <- get.var.corrected(mxFit=guillenFit, fixed.params=fixed.params, 
        V.fixed=var.err.var.scf, uncorrected = TRUE)

# Standard errors corrected for the estimation of the measurement errors
sd.corrected <- sqrt(diag(V.total$corrected))
sd.analytic <- sqrt(diag(V.total$uncorrected))

results <- OpenMx:::parameterList(guillenFit, FALSE)
results$sd.naive.analytic <- sd.analytic
results$sd.corrected <- sd.corrected
results[, 5:8] <- round(results[, 5:8], 3)

results$Est.err <- round(100 * 
  (diag(V.total$corrected) - diag(V.total$uncorrected)) / diag(V.total$corrected)
)
results <- results[, -c(1,2)]

print(results)

#sd.corrected.should.be <-
#c(0.06134, 0.05415, 0.06727, 0.04793, 0.05949, 0.16464, 0.17048, 
#    0.22521, 0.34188, 0.11237, 0.12872, 0.04943, 0.09895, 0.12774, 
#0.25775, 0.15589, 0.23150, 0.12528, 0.05655)

#stopifnot(abs(sd.corrected - sd.corrected.should.be) < 1e-5)
